@@ -148,7 +148,7 @@ struct bcm2835_host {
 	spinlock_t		lock;
 
 	void __iomem		*ioaddr;
-	u32			bus_addr;
+	phys_addr_t		bus_addr;
 
 	struct mmc_host		*mmc;
 
@@ -246,8 +246,8 @@ static void log_init(struct device *dev, u32 bus_to_phys)
 	sdhost_log_buf = dma_zalloc_coherent(dev, LOG_SIZE, &sdhost_log_addr,
 					     GFP_KERNEL);
 	if (sdhost_log_buf) {
-		pr_info("sdhost: log_buf @ %p (%x)\n",
-			sdhost_log_buf, sdhost_log_addr);
+		pr_info("sdhost: log_buf @ %p (%llx)\n",
+			sdhost_log_buf, (u64)sdhost_log_addr);
 		timer_base = ioremap_nocache(bus_to_phys + 0x7e003000, SZ_4K);
 		if (!timer_base)
 			pr_err("sdhost: failed to remap timer\n");
@@ -301,7 +301,7 @@ static void log_dump(void)
 	}
 }
 
-#define log_event(event, param1, param2) log_event_impl(event, param1, param2)
+#define log_event(event, param1, param2) log_event_impl(event, (u32)(uintptr_t)param1, (u32)(uintptr_t)param2)
 
 #else
 
@@ -527,7 +527,7 @@ static void bcm2835_sdhost_dma_complete(void *param)
 	unsigned long flags;
 
 	spin_lock_irqsave(&host->lock, flags);
-	log_event("DMA<", (u32)host->data, bcm2835_sdhost_read(host, SDHSTS));
+	log_event("DMA<", host->data, bcm2835_sdhost_read(host, SDHSTS));
 	log_event("DMA ", bcm2835_sdhost_read(host, SDCMD),
 		  bcm2835_sdhost_read(host, SDEDM));
 
@@ -543,6 +543,11 @@ static void bcm2835_sdhost_dma_complete(void *param)
 		void *page;
 		u32 *buf;
 
+		if (host->drain_offset & PAGE_MASK) {
+			host->drain_page += host->drain_offset >> PAGE_SHIFT;
+			host->drain_offset &= ~PAGE_MASK;
+		}
+
 		page = kmap_atomic(host->drain_page);
 		buf = page + host->drain_offset;
 
@@ -559,7 +564,7 @@ static void bcm2835_sdhost_dma_complete(void *param)
 
 	bcm2835_sdhost_finish_data(host);
 
-	log_event("DMA>", (u32)host->data, 0);
+	log_event("DMA>", host->data, 0);
 	spin_unlock_irqrestore(&host->lock, flags);
 }
 
@@ -748,7 +753,7 @@ static void bcm2835_sdhost_transfer_pio(struct bcm2835_host *host)
 	u32 sdhsts;
 	bool is_read;
 	BUG_ON(!host->data);
-	log_event("XFP<", (u32)host->data, host->blocks);
+	log_event("XFP<", host->data, host->blocks);
 
 	is_read = (host->data->flags & MMC_DATA_READ) != 0;
 	if (is_read)
@@ -773,7 +778,7 @@ static void bcm2835_sdhost_transfer_pio(struct bcm2835_host *host)
 		       sdhsts);
 		host->data->error = -ETIMEDOUT;
 	}
-	log_event("XFP>", (u32)host->data, host->blocks);
+	log_event("XFP>", host->data, host->blocks);
 }
 
 static void bcm2835_sdhost_prepare_dma(struct bcm2835_host *host,
@@ -783,7 +788,7 @@ static void bcm2835_sdhost_prepare_dma(struct bcm2835_host *host,
 	struct dma_async_tx_descriptor *desc = NULL;
 	struct dma_chan *dma_chan;
 
-	log_event("PRD<", (u32)data, 0);
+	log_event("PRD<", data, 0);
 	pr_debug("bcm2835_sdhost_prepare_dma()\n");
 
 	dma_chan = host->dma_chan_rxtx;
@@ -794,7 +799,7 @@ static void bcm2835_sdhost_prepare_dma(struct bcm2835_host *host,
 		dir_data = DMA_TO_DEVICE;
 		dir_slave = DMA_MEM_TO_DEV;
 	}
-	log_event("PRD1", (u32)dma_chan, 0);
+	log_event("PRD1", dma_chan, 0);
 
 	BUG_ON(!dma_chan->device);
 	BUG_ON(!dma_chan->device->dev);
@@ -841,7 +846,7 @@ static void bcm2835_sdhost_prepare_dma(struct bcm2835_host *host,
 		desc = dmaengine_prep_slave_sg(dma_chan, data->sg,
 					       len, dir_slave,
 					       DMA_PREP_INTERRUPT | DMA_CTRL_ACK);
-	log_event("PRD3", (u32)desc, 0);
+	log_event("PRD3", desc, 0);
 
 	if (desc) {
 		desc->callback = bcm2835_sdhost_dma_complete;
@@ -850,12 +855,12 @@ static void bcm2835_sdhost_prepare_dma(struct bcm2835_host *host,
 		host->dma_chan = dma_chan;
 		host->dma_dir = dir_data;
 	}
-	log_event("PDM>", (u32)data, 0);
+	log_event("PDM>", data, 0);
 }
 
 static void bcm2835_sdhost_start_dma(struct bcm2835_host *host)
 {
-	log_event("SDMA", (u32)host->data, (u32)host->dma_chan);
+	log_event("SDMA", host->data, host->dma_chan);
 	dmaengine_submit(host->dma_desc);
 	dma_async_issue_pending(host->dma_chan);
 }
@@ -1079,7 +1084,7 @@ static void bcm2835_sdhost_finish_data(struct bcm2835_host *host)
 	data = host->data;
 	BUG_ON(!data);
 
-	log_event("FDA<", (u32)host->mrq, (u32)host->cmd);
+	log_event("FDA<", host->mrq, host->cmd);
 	pr_debug("finish_data(error %d, stop %d, sbc %d)\n",
 	       data->error, data->stop ? 1 : 0,
 	       host->mrq->sbc ? 1 : 0);
@@ -1102,7 +1107,7 @@ static void bcm2835_sdhost_finish_data(struct bcm2835_host *host)
 	}
 	else
 		bcm2835_sdhost_transfer_complete(host);
-	log_event("FDA>", (u32)host->mrq, (u32)host->cmd);
+	log_event("FDA>", host->mrq, host->cmd);
 }
 
 static void bcm2835_sdhost_transfer_complete(struct bcm2835_host *host)
@@ -1116,7 +1121,7 @@ static void bcm2835_sdhost_transfer_complete(struct bcm2835_host *host)
 	data = host->data;
 	host->data = NULL;
 
-	log_event("TCM<", (u32)data, data->error);
+	log_event("TCM<", data, data->error);
 	pr_debug("transfer_complete(error %d, stop %d)\n",
 	       data->error, data->stop ? 1 : 0);
 
@@ -1138,7 +1143,7 @@ static void bcm2835_sdhost_transfer_complete(struct bcm2835_host *host)
 		bcm2835_sdhost_wait_transfer_complete(host);
 		tasklet_schedule(&host->finish_tasklet);
 	}
-	log_event("TCM>", (u32)data, 0);
+	log_event("TCM>", data, 0);
 }
 
 /* If irq_flags is valid, the caller is in a thread context and is allowed
@@ -1153,7 +1158,7 @@ static void bcm2835_sdhost_finish_command(struct bcm2835_host *host,
 	int timediff = 0;
 #endif
 
-	log_event("FCM<", (u32)host->mrq, (u32)host->cmd);
+	log_event("FCM<", host->mrq, host->cmd);
 	pr_debug("finish_command(%x)\n", bcm2835_sdhost_read(host, SDCMD));
 
 	BUG_ON(!host->cmd || !host->mrq);
@@ -1244,6 +1249,8 @@ static void bcm2835_sdhost_finish_command(struct bcm2835_host *host,
 				pr_info("%s: ignoring CRC7 error for CMD1\n",
 					mmc_hostname(host->mmc));
 		} else {
+			u32 edm, fsm;
+
 			if (sdhsts & SDHSTS_CMD_TIME_OUT) {
 				if (host->debug)
 					pr_warn("%s: command %d timeout\n",
@@ -1256,6 +1263,13 @@ static void bcm2835_sdhost_finish_command(struct bcm2835_host *host,
 				       host->cmd->opcode);
 				host->cmd->error = -EILSEQ;
 			}
+
+			edm = readl(host->ioaddr + SDEDM);
+			fsm = edm & SDEDM_FSM_MASK;
+			if (fsm == SDEDM_FSM_READWAIT ||
+			    fsm == SDEDM_FSM_WRITESTART1)
+				writel(edm | SDEDM_FORCE_DATA_MODE,
+				       host->ioaddr + SDEDM);
 			tasklet_schedule(&host->finish_tasklet);
 			return;
 		}
@@ -1301,16 +1315,14 @@ static void bcm2835_sdhost_finish_command(struct bcm2835_host *host,
 		else if (host->data_complete)
 			bcm2835_sdhost_transfer_complete(host);
 	}
-	log_event("FCM>", (u32)host->mrq, (u32)host->cmd);
+	log_event("FCM>", host->mrq, host->cmd);
 }
 
-static void bcm2835_sdhost_timeout(unsigned long data)
+static void bcm2835_sdhost_timeout(struct timer_list *t)
 {
-	struct bcm2835_host *host;
+	struct bcm2835_host *host = from_timer(host, t, timer);
 	unsigned long flags;
 
-	host = (struct bcm2835_host *)data;
-
 	spin_lock_irqsave(&host->lock, flags);
 	log_event("TIM<", 0, 0);
 
@@ -1340,7 +1352,7 @@ static void bcm2835_sdhost_timeout(unsigned long data)
 
 static void bcm2835_sdhost_busy_irq(struct bcm2835_host *host, u32 intmask)
 {
-	log_event("IRQB", (u32)host->cmd, intmask);
+	log_event("IRQB", host->cmd, intmask);
 	if (!host->cmd) {
 		pr_err("%s: got command busy interrupt 0x%08x even "
 			"though no command operation was in progress.\n",
@@ -1393,7 +1405,7 @@ static void bcm2835_sdhost_data_irq(struct bcm2835_host *host, u32 intmask)
 	   data/space available FIFO status bits. It is therefore not
 	   an error to get here when there is no data transfer in
 	   progress. */
-	log_event("IRQD", (u32)host->data, intmask);
+	log_event("IRQD", host->data, intmask);
 	if (!host->data)
 		return;
 
@@ -1430,7 +1442,7 @@ static void bcm2835_sdhost_data_irq(struct bcm2835_host *host, u32 intmask)
 
 static void bcm2835_sdhost_block_irq(struct bcm2835_host *host, u32 intmask)
 {
-	log_event("IRQK", (u32)host->data, intmask);
+	log_event("IRQK", host->data, intmask);
 	if (!host->data) {
 		pr_err("%s: got block interrupt 0x%08x even "
 			"though no data operation was in progress.\n",
@@ -1688,10 +1700,10 @@ static void bcm2835_sdhost_request(struct mmc_host *mmc, struct mmc_request *mrq
 	edm = bcm2835_sdhost_read(host, SDEDM);
 	fsm = edm & SDEDM_FSM_MASK;
 
-	log_event("REQ<", (u32)mrq, edm);
+	log_event("REQ<", mrq, edm);
 	if ((fsm != SDEDM_FSM_IDENTMODE) &&
 	    (fsm != SDEDM_FSM_DATAMODE)) {
-		log_event("REQ!", (u32)mrq, edm);
+		log_event("REQ!", mrq, edm);
 		if (host->debug) {
 			pr_warn("%s: previous command (%d) not complete (EDM %x)\n",
 			       mmc_hostname(host->mmc),
@@ -1723,11 +1735,11 @@ static void bcm2835_sdhost_request(struct mmc_host *mmc, struct mmc_request *mrq
 			bcm2835_sdhost_finish_command(host, &flags);
 	}
 
-	log_event("CMD ", (u32)mrq->cmd->opcode,
+	log_event("CMD ", mrq->cmd->opcode,
 		   mrq->data ? (u32)mrq->data->blksz : 0);
 	mmiowb();
 
-	log_event("REQ>", (u32)mrq, 0);
+	log_event("REQ>", mrq, 0);
 	spin_unlock_irqrestore(&host->lock, flags);
 }
 
@@ -1783,7 +1795,7 @@ static void bcm2835_sdhost_cmd_wait_work(struct work_struct *work)
 
 	spin_lock_irqsave(&host->lock, flags);
 
-	log_event("CWK<", (u32)host->cmd, (u32)host->mrq);
+	log_event("CWK<", host->cmd, host->mrq);
 
 	/*
 	 * If this tasklet gets rescheduled while running, it will
@@ -1798,7 +1810,7 @@ static void bcm2835_sdhost_cmd_wait_work(struct work_struct *work)
 
 	mmiowb();
 
-	log_event("CWK>", (u32)host->cmd, 0);
+	log_event("CWK>", host->cmd, 0);
 
 	spin_unlock_irqrestore(&host->lock, flags);
 }
@@ -1814,7 +1826,7 @@ static void bcm2835_sdhost_tasklet_finish(unsigned long param)
 
 	spin_lock_irqsave(&host->lock, flags);
 
-	log_event("TSK<", (u32)host->mrq, 0);
+	log_event("TSK<", host->mrq, 0);
 	/*
 	 * If this tasklet gets rescheduled while running, it will
 	 * be run again afterwards but without any active request.
@@ -1882,7 +1894,7 @@ static void bcm2835_sdhost_tasklet_finish(unsigned long param)
 	}
 
 	mmc_request_done(host->mmc, mrq);
-	log_event("TSK>", (u32)mrq, 0);
+	log_event("TSK>", mrq, 0);
 }
 
 int bcm2835_sdhost_add_host(struct bcm2835_host *host)
@@ -1971,8 +1983,7 @@ int bcm2835_sdhost_add_host(struct bcm2835_host *host)
 
 	INIT_WORK(&host->cmd_wait_wq, bcm2835_sdhost_cmd_wait_work);
 
-	setup_timer(&host->timer, bcm2835_sdhost_timeout,
-		    (unsigned long)host);
+	timer_setup(&host->timer, bcm2835_sdhost_timeout, 0);
 
 	bcm2835_sdhost_init(host, 0);
 
@@ -2013,6 +2024,7 @@ static int bcm2835_sdhost_probe(struct platform_device *pdev)
 	struct mmc_host *mmc;
 	const __be32 *addr;
 	u32 msg[3];
+	int na;
 	int ret;
 
 	pr_debug("bcm2835_sdhost_probe\n");
@@ -2036,12 +2048,13 @@ static int bcm2835_sdhost_probe(struct platform_device *pdev)
 		goto err;
 	}
 
+	na = of_n_addr_cells(node);
 	addr = of_get_address(node, 0, NULL, NULL);
 	if (!addr) {
 		dev_err(dev, "could not get DMA-register address\n");
 		return -ENODEV;
 	}
-	host->bus_addr = be32_to_cpup(addr);
+	host->bus_addr = (phys_addr_t)of_read_number(addr, na);
 	pr_debug(" - ioaddr %lx, iomem->start %lx, bus_addr %lx\n",
 		 (unsigned long)host->ioaddr,
 		 (unsigned long)iomem->start,
@@ -2143,6 +2156,8 @@ static int bcm2835_sdhost_probe(struct platform_device *pdev)
 
 err:
 	pr_debug("bcm2835_sdhost_probe -> err %d\n", ret);
+	if (host->dma_chan_rxtx)
+		dma_release_channel(host->dma_chan_rxtx);
 	mmc_free_host(mmc);
 
 	return ret;
@@ -2163,7 +2178,8 @@ static int bcm2835_sdhost_remove(struct platform_device *pdev)
 	del_timer_sync(&host->timer);
 
 	tasklet_kill(&host->finish_tasklet);
-
+	if (host->dma_chan_rxtx)
+		dma_release_channel(host->dma_chan_rxtx);
 	mmc_free_host(host->mmc);
 	platform_set_drvdata(pdev, NULL);
 
